//#include <omp.h>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <cmath>
#include "nr.h"
#include <vector>
#include <stdio.h> 
#include <time.h>
using namespace std;
#include "mpi.h"
# include "minpack.H"
# include "asa047.H"

const int N_cpu = 100; //200;

double beta = 0.96;
double theta = 0.36;
double delta = 0.08;
double alpha_p = 0.13;
double alpha_c = 0.30;
double a_h = 1.024;
double a_l = 0.976;
double a0, a0_in;
double p11, p12, p21, p22;
double pi_1, pi_2;

double tau_k, tau_l;
double tau_k_in, tau_l_in;
const int Ntau = 100;
double tau_vec[Ntau];
double tau_a = 0.001; 
double tau_b = 0.9;   //boundary of tax rate.

const int Nk = 400; //200;
const int Ns = 2;
const int Nbdr = 2;
const int Nm = 50;
const int Nm_search = 1000;
//const int Nv = 500;

double kl = 0.01;
double kh = 2.50;
double k0, k0_in, k1;
double k_vec[Nk];
double Wm_0[Nk][Ns][Nbdr], Wm_1[Nk][Ns][Nbdr];   //ind. payoff
double Wv0_0[Nk][Ns][Nm], Wv0_1[Nk][Ns][Nm];             //lower bound of govt. payoff    
double Wv_0[Nk][Ns][Nm], Wv_1[Nk][Ns][Nm];       //upper bound of govt. payoff    
double W_pent[Nk][Ns];
double W_maxmin[Ntau][Ntau][Nm], k1_maxmin[Ntau][Ntau][Nm], c_maxmin[Ntau][Ntau][Nm], l_maxmin[Ntau][Ntau][Nm];
double tmp_W_maxmin[Ntau][Ntau], tmp_k1_maxmin[Ntau][Ntau], tmp_c_maxmin[Ntau][Ntau], tmp_l_maxmin[Ntau][Ntau];
double gc_0[Nk][Ns], gc_1[Nk][Ns], gl_0[Nk][Ns], gl_1[Nk][Ns];
double gk1_0[Nk][Ns], gtauk_0[Nk][Ns], gtaul_0[Nk][Ns];
double gk1_1[Nk][Ns], gtauk_1[Nk][Ns], gtaul_1[Nk][Ns];
double renew_v[Nbdr];
double v_lower = -1.15;
double v_upper = -0.75;

double m0_in;
double m0, m1;
double check_m, check_v;
double check_vec[2];

double out_m_k[7];  //k1, l, uc, u, ff, c0, g

//update v:
int vl_shk;
double vl_k0, vh_k1, vh_v0;
double m0_s1_l[Nm], m0_s1_r[Nm];
double m1_s1_l[Nm], m1_s1_r[Nm];
double m1_s2_l[Nm], m1_s2_r[Nm];

double tol_Wm = 1e-4;  //1e-8; 
double tol_f_mk = 1e-10;
double tol_v0;

double max_dis_m, max_dis_vl, max_dis_vh;

//***********************************************************

void NRlocate(double xx[], int size, double x, int &j)
{
	int ju,jm,jl;
	bool ascnd;

	int n=size;
	jl=-1;
	ju=n;
	ascnd=(xx[n-1] >= xx[0]);
	while (ju-jl > 1) {
		jm=(ju+jl) >> 1;
		if (x >= xx[jm] == ascnd)
			jl=jm;
		else
			ju=jm;
	}
	if (x == xx[0]) j=0;
	else if (x == xx[n-1]) j=n-2;
	else j=jl;
}

bool check() {
        bool ans = true;

		for (int row = 0; row < Nk; row++) {
			for (int i_s = 0; i_s < Ns; i_s++) {
				for (int i_m = 0; i_m < Nbdr; i_m++) {
					if ( fabs (Wm_0[row][i_s][i_m] - Wm_1[row][i_s][i_m]) > tol_Wm ){
						return false;
					}
				}
			}
		}

    return ans;
}

void check_dis() 
{
	max_dis_m = -100;
	max_dis_vl = -100;
	max_dis_vh = -100;

        for (int row = 0; row < Nk; row++) {
			for (int i_s = 0; i_s < Ns; i_s++) {

				for (int i_m = 0; i_m < Nbdr; i_m++) {
					if ( fabs (Wm_0[row][i_s][i_m] - Wm_1[row][i_s][i_m]) > max_dis_m ){
						max_dis_m = fabs (Wm_0[row][i_s][i_m] - Wm_1[row][i_s][i_m]);
					} 
				}
				
				for (int i_m = 0; i_m < Nm; i_m++) {
                     
					if ( fabs (Wv0_0[row][i_s][i_m] - Wv0_1[row][i_s][i_m]) > max_dis_vl ){
						max_dis_vl = fabs (Wv0_0[row][i_s][i_m] - Wv0_1[row][i_s][i_m]);
					}
					
					if ( fabs (Wv_0[row][i_s][i_m] - Wv_1[row][i_s][i_m]) > max_dis_vh ){
						max_dis_vh = fabs (Wv_0[row][i_s][i_m] - Wv_1[row][i_s][i_m]);
					}
				}
			}
		}
}

void shock(int &shk, double x_a, double &x_p1, double &x_p2)
{
	if (shk == 0) {
		x_a = a_l;
		x_p1 = p11;
		x_p2 = p12;
	}
	else if (shk == 1) {
		x_a = a_h;
		x_p1 = p21;
		x_p2 = p22;
	}
}

double f_m_k_brent (DP x)
{
	double x_la;   
	double pent = 0.0;
	
	x_la = x; 
	if (x_la < 0.001){                    
		pent += 100*(x_la-0.001)*(x_la-0.001);
		x_la = 0.001; 
	} 
	if (x_la > 0.999){                    
		pent += 100*(x_la-0.999)*(x_la-0.999);
		x_la = 0.999; 
	}
	
	//m_in;
	double x_w0, x_r0, x_c0, x_uc, x_ul;
	x_w0 = a0_in*(1.0-theta)*pow(k0_in/x_la, theta);
	x_r0 = a0_in*theta*pow(k0_in/x_la, theta-1.0);

	double tmp1, tmp2;
	tmp1 = m0_in*(1.0-tau_l_in)*x_w0*(1-x_la);
	tmp2 = (1.0-beta)*(1.0-alpha_p)*(1.0-alpha_c)*( 1.0+(1.0-tau_k_in)*(x_r0-delta) );

	return fabs(tmp1/tmp2 - 1.0) + pent;
} 

void m_k(double &x_k0, int &x_shk1, double &x_m0, double &x_tau_k, double &x_tau_l)
{
	if (x_shk1 == 0){
		a0_in = a_l;
	}
	if (x_shk1 == 1){
		a0_in = a_h;
	}
	tau_k_in = x_tau_k;
	tau_l_in = x_tau_l;
	m0_in = x_m0;
	k0_in = x_k0;
	                                       
	const DP TOL=1.0e-12,EQL=1.0e-10;
	bool newroot;
	int j;
	int nmin=0;
	DP ax, bx, cx;
	DP xmin1, xmin2;
	Vec_DP amin(60);
	
	ax=0.01;
	bx=0.33;
	cx=0.99;
	
	NR::brent(ax,bx,cx,f_m_k_brent,TOL,xmin1);
	if (nmin == 0) {
		amin[nmin++]=xmin1;
	} else {
		newroot=true;
		for (j=0;j<nmin;j++){
			if (fabs(xmin1-amin[j]) <= (EQL*xmin1)) newroot=false;
			if (newroot) {
				amin[nmin++]=xmin1;
			}
		}
	}  

	double x_w0, x_r0, x_c0, x_uc, x_ul, x_u0, x_g, x_ug;
	double x_k1, x_la;
	
	x_la = xmin1;  

	x_w0 = a0_in*(1.0-theta)*pow(k0_in/x_la, theta);
	x_r0 = a0_in*theta*pow(k0_in/x_la, theta-1.0);

	x_c0 = alpha_c*(1.0-tau_l_in)*x_w0*(1.0-x_la)/(1.0-alpha_c);

	x_k1 = (1.0-tau_l_in)*x_w0*x_la + (1.0-tau_k_in)*x_r0*k0_in + tau_k_in*delta*k0_in + (1.0-delta)*k0_in - x_c0;

	x_uc = (1.0-beta)*(1.0-alpha_p)*alpha_c/x_c0;
	x_u0 = (1.0-beta)* ( (1.0-alpha_p)*alpha_c*log(x_c0) + (1.0-alpha_p)*(1.0-alpha_c)*log(1.0-x_la) );
	x_g = tau_l_in*x_w0*x_la + tau_k_in*x_r0*k0_in - tau_k_in*delta*k0_in;
	x_ug = (1.0-beta)*alpha_p*log(x_g);

	out_m_k[0] = x_k1;             //k+
	out_m_k[1] = x_la;             //l
	out_m_k[2] = x_uc;                //uc
	out_m_k[3] = x_u0 + x_ug;         //u
	out_m_k[4] = f_m_k_brent(xmin1);     //fvec
	out_m_k[5] = x_c0;
	out_m_k[6] = x_g;

}
 
double intp_2d(double &x_k1, double &x_m1, int &shk, 
	double m_vec_l[Nm], double m_vec_r[Nm], double policy[Nk][Ns][Nm])
{   
	int i_k;	
	NRlocate(k_vec, Nk, x_k1, i_k);
    if (i_k==-1){i_k = 0;}
	if (i_k==Nk-1){i_k = Nk-2;}

	int lef1, lef2;
	NRlocate(m_vec_l, Nm, x_m1, lef1);
	if(lef1<0){lef1=0;}
	if(lef1>Nm-2){lef1=Nm-2;}

	NRlocate(m_vec_r, Nm, x_m1, lef2);
	if(lef2<0){lef2=0;}
	if(lef2>Nm-2){lef2=Nm-2;}

	double fa, fb, fc, fd, fe, ff;

	fa = policy[i_k][shk][lef1+1];
	fb = policy[i_k+1][shk][lef2+1];
	fc = policy[i_k+1][shk][lef2];
	fd = policy[i_k][shk][lef1];

	fe = fd + (fa-fd)/(m_vec_l[lef1+1] - m_vec_l[lef1])*(x_m1 - m_vec_l[lef1]);
	ff = fc + (fb-fc)/(m_vec_r[lef2+1] - m_vec_r[lef2])*(x_m1 - m_vec_r[lef2]);
		
	return fe + (ff-fe)/(k_vec[i_k+1] - k_vec[i_k])*(x_k1 - k_vec[i_k]);	
} 

void check_equ(double &x_k0, int &x_shk1, double &x_m0)
{
	double x_k1, x_l, x_uc, x_u;
	double m11_a, m11_b, m12_a, m12_b; 
	double x_m11, x_m12;
	int lef, i_m11, i_m12;
	double tmp_m1_l, tmp_m1_r;
	double x_v0, x_v11, x_v12;
	double m1_vec_l[Nm], m1_vec_r[Nm];
	int lef_m;
	int x_shk2;

	shock(x_shk1, a0, pi_1, pi_2);   
	
	NRlocate(k_vec, Nk, x_k0, lef);
	if(lef<0){lef=0;}
	if(lef>Nk-2){lef=Nk-2;}

	//lower bound, payoff of govt. detm. by k0, s0.
	for (int i_m = 0; i_m < Nm; i_m++){
		m1_vec_l[i_m] = (Wm_0[lef][x_shk1][1] - Wm_0[lef][x_shk1][0])/(Nm-1) * i_m + Wm_0[lef][x_shk1][0];
		m1_vec_r[i_m] = (Wm_0[lef+1][x_shk1][1] - Wm_0[lef+1][x_shk1][0])/(Nm-1) * i_m + Wm_0[lef+1][x_shk1][0];
	}      
	
	//x_v0 = intp_2d(x_k0, x_m0, x_shk1, m1_vec_l, m1_vec_r, Wv0_0);	
	x_v0 = W_pent[lef][x_shk1];  
	
	check_vec[0] = 0.0;
	check_vec[1] = 0.0;

	int i_k, i_l;
	i_k = -1;
	
	do{
		i_k += 1;
		tau_k = tau_vec[i_k];
		   
		i_l = -1;
	    do{
			i_l += 1;
			tau_l = tau_vec[i_l];

			m_k(x_k0, x_shk1, x_m0, tau_k, tau_l);
			x_k1 = out_m_k[0];
			x_l = out_m_k[1];
			x_uc = out_m_k[2];
			x_u = out_m_k[3];
		               
			if ( out_m_k[4] <= tol_f_mk && x_k1 >= (1.0-delta)*x_k0 
						&& x_k1 <= kh && x_l > 0.0 && x_l < 1.0
						&& out_m_k[6] > 0 ){  
			
				NRlocate(k_vec, Nk, x_k1, lef);
				if(lef<0){lef=0;}
				if(lef>Nk-2){lef=Nk-2;}  
			   
			//temp commented.
			    for (int i_m = 0; i_m < Nm; i_m++){
					m1_s1_l[i_m] = (Wm_0[lef][0][1] - Wm_0[lef][0][0])/(Nm-1) * i_m + Wm_0[lef][0][0];
					m1_s1_r[i_m] = (Wm_0[lef+1][0][1] - Wm_0[lef+1][0][0])/(Nm-1) * i_m + Wm_0[lef+1][0][0];
				}

				for (int i_m = 0; i_m < Nm; i_m++){
					m1_s2_l[i_m] = (Wm_0[lef][1][1] - Wm_0[lef][1][0])/(Nm-1) * i_m + Wm_0[lef][1][0];
					m1_s2_r[i_m] = (Wm_0[lef+1][1][1] - Wm_0[lef+1][1][0])/(Nm-1) * i_m + Wm_0[lef+1][1][0];
				}  
	
				i_m11 = -1;
				do{
					i_m11 += 1;

					//payoff of ind. shk+ = 0
					tmp_m1_l = (Wm_0[lef][0][1] - Wm_0[lef][0][0])/(Nm_search-1) * i_m11 + Wm_0[lef][0][0];
					tmp_m1_r = (Wm_0[lef+1][0][1] - Wm_0[lef+1][0][0])/(Nm_search-1) * i_m11 + Wm_0[lef+1][0][0];  
					x_m11 = (tmp_m1_r-tmp_m1_l)/(k_vec[lef+1]-k_vec[lef])*(x_k1-k_vec[lef]) + tmp_m1_l;  
				    
				//temp commented.                                                                               
					x_shk2 = 0;
					x_v11 = intp_2d(x_k1, x_m11, x_shk2, m1_s1_l, m1_s1_r, Wv_0);   

					x_m12 = (x_uc/beta - pi_1*x_m11) / pi_2;

				//temp commented.                                                                            
					x_shk2 = 1;
					x_v12 = intp_2d(x_k1, x_m12, x_shk1, m1_s2_l, m1_s2_r, Wv_0); 

					//test:
						if ( x_m12 >= Wm_0[lef][1][0] && x_m12 <= Wm_0[lef][1][1] 
						      && x_u + beta * (pi_1*x_v11 + pi_2*x_v12) >= x_v0
							) 
						{
							check_vec[0] = 1.0;
						}

						if ( x_m12 >= Wm_0[lef+1][1][0] && x_m12 <= Wm_0[lef+1][1][1] 
						      && x_u + beta * (pi_1*x_v11 + pi_2*x_v12) >= x_v0
							) 
						{
							check_vec[0] = 1.0;
						}
						//end test.
			
				} while (i_m11 < Nm_search &&  check_vec[0] < 1.0); //end m11 
			
			 } //end if out_m_k[4] <= tol_f_mk

		} while (tau_l <= tau_b && check_vec[0] < 1.0); //end tau_k, tau_l
	}while (tau_k <= tau_b && check_vec[0] < 1.0); //end tau_k, tau_l
}

double max_dp(double &x1, double &x2)
{
	double tmp;
	if (x1>=x2){
		tmp = x1;
	}
	if (x2>=x1){
		tmp = x2;
	}
	return tmp;
}

double min_dp(double &x1, double &x2)
{
	double tmp;
	if (x1>=x2){
		tmp = x2;
	}
	if (x2>=x1){
		tmp = x1;
	}
	return tmp;
}

//search over m+(1)
double eval_f_vh1(DP x)
{
	double x_m11, x_m12;
	
	x_m11 = x;
	x_m12 =	( out_m_k[2]/beta - pi_1*x_m11 ) / pi_2;

	double pent = 0.0;
         
	if (x_m11 < m1_s1_l[0]){                    
		pent += 100*(x_m11- m1_s1_l[0])*(x_m11- m1_s1_l[0]);
		x_m11 = m1_s1_l[0]; 
	} 
	if (x_m11 > m1_s1_l[Nm-1]){                    
		pent += 100*(x_m11-m1_s1_l[Nm-1])*(x_m11-m1_s1_l[Nm-1]);
		x_m11 = m1_s1_l[Nm-1]; 
	}
	//*************************
	if (x_m12 < m1_s2_l[0]){                    
		pent += 100*(x_m12- m1_s2_l[0])*(x_m12- m1_s2_l[0]);
		x_m12 = m1_s2_l[0]; 
	} 
	if (x_m12 > m1_s2_l[Nm-1]){                    
		pent += 100*(x_m12-m1_s2_l[Nm-1])*(x_m12-m1_s2_l[Nm-1]);
		x_m12 = m1_s2_l[Nm-1]; 
	}
	 
	double ev1, ev2;
	int x_shk1;
	x_shk1 = 0;
	ev1 = intp_2d(vh_k1, x_m11, x_shk1, m1_s1_l, m1_s1_r, Wv_0);

	x_shk1 = 1;
	ev2 = intp_2d(vh_k1, x_m12, x_shk1, m1_s2_l, m1_s2_r, Wv_0);
	
	double tmp_f;
	tmp_f = -( out_m_k[3] + beta * (pi_1*ev1 + pi_2*ev2) );

	return tmp_f + pent;
}

double eval_f_vh1_npent(DP x)
{
	double x_m11, x_m12;
	
	x_m11 = x;
	x_m12 =	( out_m_k[2]/beta - pi_1*x_m11 ) / pi_2;

	double pent = 0.0;
         
	if (x_m11 < m1_s1_l[0]){                    
		//pent += 100*(x_m11- m1_s1_l[0])*(x_m11- m1_s1_l[0]);
		x_m11 = m1_s1_l[0]; 
	} 
	if (x_m11 > m1_s1_l[Nm-1]){                    
		//pent += 100*(x_m11-m1_s1_l[Nm-1])*(x_m11-m1_s1_l[Nm-1]);
		x_m11 = m1_s1_l[Nm-1]; 
	}
	//*************************
	if (x_m12 < m1_s2_l[0]){                    
		//pent += 100*(x_m12- m1_s2_l[0])*(x_m12- m1_s2_l[0]);
		x_m12 = m1_s2_l[0]; 
	} 
	if (x_m12 > m1_s2_l[Nm-1]){                    
		//pent += 100*(x_m12-m1_s2_l[Nm-1])*(x_m12-m1_s2_l[Nm-1]);
		x_m12 = m1_s2_l[Nm-1]; 
	}
	 
	double ev1, ev2;
	int x_shk1;
	x_shk1 = 0;
	ev1 = intp_2d(vh_k1, x_m11, x_shk1, m1_s1_l, m1_s1_r, Wv_0);

	x_shk1 = 1;
	ev2 = intp_2d(vh_k1, x_m12, x_shk1, m1_s2_l, m1_s2_r, Wv_0);
	
	double tmp_f;
	tmp_f = -( out_m_k[3] + beta * (pi_1*ev1 + pi_2*ev2) );

	return tmp_f;
}

//search over m+(2)
double eval_f_vh2(DP x)
{
	double x_m11, x_m12;
	
	x_m12 = x;
	x_m11 =	( out_m_k[2]/beta - pi_2*x_m12 ) / pi_1;

	double pent = 0.0;
         
	if (x_m11 < m1_s1_l[0]){                    
		pent += 100*(x_m11- m1_s1_l[0])*(x_m11- m1_s1_l[0]);
		x_m11 = m1_s1_l[0]; 
	} 
	if (x_m11 > m1_s1_l[Nm-1]){                    
		pent += 100*(x_m11-m1_s1_l[Nm-1])*(x_m11-m1_s1_l[Nm-1]);
		x_m11 = m1_s1_l[Nm-1]; 
	}
	//*************************
	if (x_m12 < m1_s2_l[0]){                    
		pent += 100*(x_m12- m1_s2_l[0])*(x_m12- m1_s2_l[0]);
		x_m12 = m1_s2_l[0]; 
	} 
	if (x_m12 > m1_s2_l[Nm-1]){                    
		pent += 100*(x_m12-m1_s2_l[Nm-1])*(x_m12-m1_s2_l[Nm-1]);
		x_m12 = m1_s2_l[Nm-1]; 
	}
	 
	double ev1, ev2;
	int x_shk1;
	x_shk1 = 0;
	ev1 = intp_2d(vh_k1, x_m11, x_shk1, m1_s1_l, m1_s1_r, Wv_0);

	x_shk1 = 1;
	ev2 = intp_2d(vh_k1, x_m12, x_shk1, m1_s2_l, m1_s2_r, Wv_0);
	
	double tmp_f;
	tmp_f = -( out_m_k[3] + beta * (pi_1*ev1 + pi_2*ev2) );

	return tmp_f + pent;
}

double eval_f_vh2_npent(DP x)
{
	double x_m11, x_m12;
	
	x_m12 = x;
	x_m11 =	( out_m_k[2]/beta - pi_2*x_m12 ) / pi_1;

	double pent = 0.0;
         
	if (x_m11 < m1_s1_l[0]){                    
		//pent += 100*(x_m11- m1_s1_l[0])*(x_m11- m1_s1_l[0]);
		x_m11 = m1_s1_l[0]; 
	} 
	if (x_m11 > m1_s1_l[Nm-1]){                    
		//pent += 100*(x_m11-m1_s1_l[Nm-1])*(x_m11-m1_s1_l[Nm-1]);
		x_m11 = m1_s1_l[Nm-1]; 
	}
	//*************************
	if (x_m12 < m1_s2_l[0]){                    
		//pent += 100*(x_m12- m1_s2_l[0])*(x_m12- m1_s2_l[0]);
		x_m12 = m1_s2_l[0]; 
	} 
	if (x_m12 > m1_s2_l[Nm-1]){                    
		//pent += 100*(x_m12-m1_s2_l[Nm-1])*(x_m12-m1_s2_l[Nm-1]);
		x_m12 = m1_s2_l[Nm-1]; 
	}
	 
	double ev1, ev2;
	int x_shk1;
	x_shk1 = 0;
	ev1 = intp_2d(vh_k1, x_m11, x_shk1, m1_s1_l, m1_s1_r, Wv_0);

	x_shk1 = 1;
	ev2 = intp_2d(vh_k1, x_m12, x_shk1, m1_s2_l, m1_s2_r, Wv_0);
	
	double tmp_f;
	tmp_f = -( out_m_k[3] + beta * (pi_1*ev1 + pi_2*ev2) );

	return tmp_f;
}
 

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//search over m+(1)
double eval_f_vl1(DP x)
{
	double x_m11, x_m12;
	
	x_m11 = x;
	x_m12 =	( out_m_k[2]/beta - pi_1*x_m11 ) / pi_2;

	double pent = 0.0;
         
	if (x_m11 < m1_s1_l[0]){                    
		pent += 100*(x_m11- m1_s1_l[0])*(x_m11- m1_s1_l[0]);
		x_m11 = m1_s1_l[0]; 
	} 
	if (x_m11 > m1_s1_l[Nm-1]){                    
		pent += 100*(x_m11-m1_s1_l[Nm-1])*(x_m11-m1_s1_l[Nm-1]);
		x_m11 = m1_s1_l[Nm-1]; 
	}
	//*************************
	if (x_m12 < m1_s2_l[0]){                    
		pent += 100*(x_m12- m1_s2_l[0])*(x_m12- m1_s2_l[0]);
		x_m12 = m1_s2_l[0]; 
	} 
	if (x_m12 > m1_s2_l[Nm-1]){                    
		pent += 100*(x_m12-m1_s2_l[Nm-1])*(x_m12-m1_s2_l[Nm-1]);
		x_m12 = m1_s2_l[Nm-1]; 
	}
	 
	double ev1, ev2;
	int x_shk1;
	x_shk1 = 0;
	ev1 = intp_2d(vh_k1, x_m11, x_shk1, m1_s1_l, m1_s1_r, Wv0_0);

	x_shk1 = 1;
	ev2 = intp_2d(vh_k1, x_m12, x_shk1, m1_s2_l, m1_s2_r, Wv0_0);
	
	double tmp_f;
	tmp_f = ( out_m_k[3] + beta * (pi_1*ev1 + pi_2*ev2) );

	return tmp_f + pent;
}

double eval_f_vl1_npent(DP x)
{
	double x_m11, x_m12;
	
	x_m11 = x;
	x_m12 =	( out_m_k[2]/beta - pi_1*x_m11 ) / pi_2;

	double pent = 0.0;
         
	if (x_m11 < m1_s1_l[0]){                    
		//pent += 100*(x_m11- m1_s1_l[0])*(x_m11- m1_s1_l[0]);
		x_m11 = m1_s1_l[0]; 
	} 
	if (x_m11 > m1_s1_l[Nm-1]){                    
		//pent += 100*(x_m11-m1_s1_l[Nm-1])*(x_m11-m1_s1_l[Nm-1]);
		x_m11 = m1_s1_l[Nm-1]; 
	}
	//*************************
	if (x_m12 < m1_s2_l[0]){                    
		//pent += 100*(x_m12- m1_s2_l[0])*(x_m12- m1_s2_l[0]);
		x_m12 = m1_s2_l[0]; 
	} 
	if (x_m12 > m1_s2_l[Nm-1]){                    
		//pent += 100*(x_m12-m1_s2_l[Nm-1])*(x_m12-m1_s2_l[Nm-1]);
		x_m12 = m1_s2_l[Nm-1]; 
	}
	 
	double ev1, ev2;
	int x_shk1;
	x_shk1 = 0;
	ev1 = intp_2d(vh_k1, x_m11, x_shk1, m1_s1_l, m1_s1_r, Wv0_0);

	x_shk1 = 1;
	ev2 = intp_2d(vh_k1, x_m12, x_shk1, m1_s2_l, m1_s2_r, Wv0_0);
	
	double tmp_f;
	tmp_f = ( out_m_k[3] + beta * (pi_1*ev1 + pi_2*ev2) );

	return tmp_f;
}

//search over m+(2)
double eval_f_vl2(DP x)
{
	double x_m11, x_m12;
	
	x_m12 = x;
	x_m11 =	( out_m_k[2]/beta - pi_2*x_m12 ) / pi_1;

	double pent = 0.0;
         
	if (x_m11 < m1_s1_l[0]){                    
		pent += 100*(x_m11- m1_s1_l[0])*(x_m11- m1_s1_l[0]);
		x_m11 = m1_s1_l[0]; 
	} 
	if (x_m11 > m1_s1_l[Nm-1]){                    
		pent += 100*(x_m11-m1_s1_l[Nm-1])*(x_m11-m1_s1_l[Nm-1]);
		x_m11 = m1_s1_l[Nm-1]; 
	}
	//*************************
	if (x_m12 < m1_s2_l[0]){                    
		pent += 100*(x_m12- m1_s2_l[0])*(x_m12- m1_s2_l[0]);
		x_m12 = m1_s2_l[0]; 
	} 
	if (x_m12 > m1_s2_l[Nm-1]){                    
		pent += 100*(x_m12-m1_s2_l[Nm-1])*(x_m12-m1_s2_l[Nm-1]);
		x_m12 = m1_s2_l[Nm-1]; 
	}
	 
	double ev1, ev2;
	int x_shk1;
	x_shk1 = 0;
	ev1 = intp_2d(vh_k1, x_m11, x_shk1, m1_s1_l, m1_s1_r, Wv0_0);

	x_shk1 = 1;
	ev2 = intp_2d(vh_k1, x_m12, x_shk1, m1_s2_l, m1_s2_r, Wv0_0);
	
	double tmp_f;
	tmp_f = ( out_m_k[3] + beta * (pi_1*ev1 + pi_2*ev2) );

	return tmp_f + pent;
}

double eval_f_vl2_npent(DP x)
{
	double x_m11, x_m12;
	
	x_m12 = x;
	x_m11 =	( out_m_k[2]/beta - pi_2*x_m12 ) / pi_1;

	double pent = 0.0;
         
	if (x_m11 < m1_s1_l[0]){                    
		x_m11 = m1_s1_l[0]; 
	} 
	if (x_m11 > m1_s1_l[Nm-1]){                    
		x_m11 = m1_s1_l[Nm-1]; 
	}
	//*************************
	if (x_m12 < m1_s2_l[0]){                    
		x_m12 = m1_s2_l[0]; 
	} 
	if (x_m12 > m1_s2_l[Nm-1]){                    
		x_m12 = m1_s2_l[Nm-1]; 
	}
	 
	double ev1, ev2;
	int x_shk1;
	x_shk1 = 0;
	ev1 = intp_2d(vh_k1, x_m11, x_shk1, m1_s1_l, m1_s1_r, Wv0_0);

	x_shk1 = 1;
	ev2 = intp_2d(vh_k1, x_m12, x_shk1, m1_s2_l, m1_s2_r, Wv0_0);
	
	double tmp_f;
	tmp_f = ( out_m_k[3] + beta * (pi_1*ev1 + pi_2*ev2) );

	return tmp_f;
}

void update_vh(double &x_k0, double x_m0, int &x_shk)
{
	double x_k1, x_l, tmp_vh, tmp_vl;  
	double tmp_vh1, tmp_vh2, tmp_vl1, tmp_vl2; 
	double x_m1, ev1_l, ev1_h;
	double m1_vec_l[Nm], m1_vec_r[Nm];
	double tmp_eval_vl, tmp_eval_vh;
	
	double tmp_x_m11, tmp_x_m12;

	tmp_vl = 1e3;
	tmp_vh = -1e3;
	tmp_vh1 = -1e3; 
	tmp_vh2 = -1e3;
	tmp_vl1 = -1e3;
	tmp_vl2 = -1e3;
	int lef;
	double vh_v1;   
	
	renew_v[0] = 0.0;
	renew_v[1] = 0.0;

	NRlocate(k_vec, Nk, x_k0, lef);
	if(lef<0){lef=0;}
	if(lef>Nk-2){lef=Nk-2;}

	//************
	//upper bound, determined by Wv_0.
	for (int i_m1 = 0; i_m1 < Nm; i_m1 ++){
		m1_vec_l[i_m1] = Wm_0[lef][x_shk][0] + 
			(Wm_0[lef][x_shk][1] - Wm_0[lef][x_shk][0]) *i_m1 / (Nm-1);
		m1_vec_r[i_m1] = Wm_0[lef+1][x_shk][0] + 
			(Wm_0[lef+1][x_shk][1] - Wm_0[lef+1][x_shk][0]) *i_m1 / (Nm-1);
	}
	//x_shk = 0;  
	
	vh_v0 = intp_2d(x_k0, x_m0, x_shk, m1_vec_l, m1_vec_r, Wv0_0);
	vh_v1 = intp_2d(x_k0, x_m0, x_shk, m1_vec_l, m1_vec_r, Wv_0);
	//************

	shock(x_shk, a0, pi_1, pi_2); 

	for (int i_tau1 = 0; i_tau1 < Ntau; i_tau1 ++){

		tau_k = tau_vec[i_tau1];
		
		for (int i_tau2 = 0; i_tau2 < Ntau; i_tau2 ++){ 
			 
			tau_l = tau_vec[i_tau2];
            
			tmp_W_maxmin[i_tau1][i_tau2] = 100.0;  //big mistake to set this equal to vh_v0.  11/11/2013

			tmp_k1_maxmin[i_tau1][i_tau2] = (1.0-delta)*x_k0;
			tmp_l_maxmin[i_tau1][i_tau2] = 0.3;
			tmp_c_maxmin[i_tau1][i_tau2] = x_k0;

			//x_shk = 0;
			m_k(x_k0, x_shk, x_m0, tau_k, tau_l);

			//*****************************//
			x_k1 = out_m_k[0];   
			x_l = out_m_k[1];
			vh_k1 = x_k1;
	                                                                                                    
			NRlocate(k_vec, Nk, x_k1, lef);
			if(lef<0){lef=0;}
			if(lef>Nk-2){lef=Nk-2;}
	
			     
			//??????
			for (int i_m1 = 0; i_m1 < Nm; i_m1 ++){
				m1_s1_l[i_m1] = Wm_0[lef][0][0] + 
					(Wm_0[lef][0][1] - Wm_0[lef][0][0]) *i_m1 / (Nm-1);
				m1_s1_r[i_m1] = Wm_0[lef+1][0][0] + 
					(Wm_0[lef+1][0][1] - Wm_0[lef+1][0][0]) *i_m1 / (Nm-1);
				//************
				m1_s2_l[i_m1] = Wm_0[lef][1][0] + 
					(Wm_0[lef][1][1] - Wm_0[lef][1][0]) *i_m1 / (Nm-1);
				m1_s2_r[i_m1] = Wm_0[lef+1][1][0] + 
					(Wm_0[lef+1][1][1] - Wm_0[lef+1][1][0]) *i_m1 / (Nm-1);
			}

		    if ( out_m_k[4] <= tol_f_mk / 1.0 && x_k1 >= (1.0-delta)*x_k0 && x_k1 <= kh && x_l > 0 && x_l < 1.0
			   //  && x_m1 >= Wm_0[lef][0][0] && x_m1 <= Wm_0[lef][0][1]
				 && out_m_k[5] > 0){ 

				tmp_k1_maxmin[i_tau1][i_tau2] = x_k1;
				tmp_c_maxmin[i_tau1][i_tau2] = out_m_k[5];
				tmp_l_maxmin[i_tau1][i_tau2] = out_m_k[1];
				
				//|||||||||||||||||||||||||||||||||||||||||||||||||||||||||
				//solve for the max {u+g+beta*Ev}
				//|||||||||||||||||||||||||||||||||||||||||||||||||||||||||

				//tmp_vh1 = -1e3; 
				//tmp_vh2 = -1e3;
				//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
				//search m+(1)
				//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
				const DP TOL=1.0e-12,EQL=1.0e-10;
				bool newroot;
				int j;
				int nmin=0;
				DP ax, bx, cx;
				DP xmin1, xmin2;
				Vec_DP amin(60);
	
				ax = (m1_s1_l[0]+m1_s1_r[0])/2.0;
				bx = (m1_s1_l[0]+m1_s1_l[Nm-1])/2.0;
				cx = (m1_s1_l[Nm-1]+m1_s1_r[Nm-1])/2.0;
	
				NR::brent(ax,bx,cx,eval_f_vh1,TOL,xmin1);
				if (nmin == 0) {
					amin[nmin++]=xmin1;
				} else {
					newroot=true;
					for (j=0;j<nmin;j++){
						if (fabs(xmin1-amin[j]) <= (EQL*xmin1)) newroot=false;
						if (newroot) {
							amin[nmin++]=xmin1;
						}
					}
				} 
				
				tmp_eval_vh = -eval_f_vh1_npent(xmin1);
				
				tmp_x_m11 = xmin1;
				tmp_x_m12 = ( out_m_k[2]/beta - pi_1*tmp_x_m11 ) / pi_2;
				
				if ( tmp_eval_vh > tmp_vh1 
					&& tmp_eval_vh >= vh_v0
					&& tmp_eval_vh <= v_upper  //vh_v1 + 0.0
					//&& tmp_x_m12 >= m1_s2_l[0]
					//&& tmp_x_m12 <= m1_s2_l[Nm-1]
                                         ){   
						
						tmp_vh1 = tmp_eval_vh;
				} 
				//end search m+(1)
				//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

				//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
				//search m+(2)
				//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
				
				ax = (m1_s2_l[0]+m1_s2_r[0])/2.0;
				bx = (m1_s2_l[0]+m1_s2_l[Nm-1])/2.0;
				cx = (m1_s2_l[Nm-1]+m1_s2_r[Nm-1])/2.0;
	
				NR::brent(ax,bx,cx,eval_f_vh2,TOL,xmin1);
				if (nmin == 0) {
					amin[nmin++]=xmin1;
				} else {
					newroot=true;
					for (j=0;j<nmin;j++){
						if (fabs(xmin1-amin[j]) <= (EQL*xmin1)) newroot=false;
						if (newroot) {
							amin[nmin++]=xmin1;
						}
					}
				} 
				//end solve 

				tmp_eval_vh = -eval_f_vh2_npent(xmin1);
				
				tmp_x_m12 = xmin1;
				tmp_x_m11 = ( out_m_k[2]/beta - pi_2*tmp_x_m12 ) / pi_1;
				
				if ( tmp_eval_vh > tmp_vh2 
					&& tmp_eval_vh >= vh_v0
					&& tmp_eval_vh <= v_upper  //vh_v1 + 0.0
					//&& tmp_x_m11 >= m1_s1_l[0]
					//&& tmp_x_m11 <= m1_s1_l[Nm-1]
                                        ){   
						
						tmp_vh2 = tmp_eval_vh;
				} 
				
				if (max_dp(tmp_vh1, tmp_vh2) >= tmp_vh){
					tmp_vh = max_dp(tmp_vh1, tmp_vh2);
				}
				//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

				//|||||||||||||||||||||||||||||||||||||||||||||||||||||||||
				//solve for the min {u+g+beta*Ev}
				//|||||||||||||||||||||||||||||||||||||||||||||||||||||||||
				
				tmp_vl1 = vh_v0; //need to initialize 
				tmp_vl2 = vh_v0;
				
				//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
				//search m+(1)
				//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
				ax = (m1_s1_l[0]+m1_s1_r[0])/2.0;
				bx = (m1_s1_l[0]+m1_s1_l[Nm-1])/2.0;
				cx = (m1_s1_l[Nm-1]+m1_s1_r[Nm-1])/2.0;

				NR::brent(ax,bx,cx,eval_f_vl1,TOL,xmin1);
				if (nmin == 0) {
					amin[nmin++]=xmin1;
				} else {
					newroot=true;
					for (j=0;j<nmin;j++){
						if (fabs(xmin1-amin[j]) <= (EQL*xmin1)) newroot=false;
						if (newroot) {
							amin[nmin++]=xmin1;
						}
					}
				} 
				//end solve 

				tmp_eval_vl = eval_f_vl1_npent(xmin1);
				
				tmp_x_m11 = xmin1;
				tmp_x_m12 = ( out_m_k[2]/beta - pi_1*tmp_x_m11 ) / pi_2;
				
				if (tmp_eval_vl >= vh_v0
					&& tmp_eval_vl <= vh_v1 + 0.0
					//&& tmp_x_m12 >= (m1_s2_l[0]+m1_s2_r[0])/2.0
					//&& tmp_x_m12 <= (m1_s2_l[Nm-1]+m1_s2_r[Nm-1])/2.0 
					){   
						
						tmp_vl1 = tmp_eval_vl;
				}
				//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

				//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
				//search m+(2)
				//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
				ax = (m1_s2_l[0]+m1_s2_r[0])/2.0;
				bx = (m1_s2_l[0]+m1_s2_l[Nm-1])/2.0;
				cx = (m1_s2_l[Nm-1]+m1_s2_r[Nm-1])/2.0;

				NR::brent(ax,bx,cx,eval_f_vl2,TOL,xmin1);
				if (nmin == 0) {
					amin[nmin++]=xmin1;
				} else {
					newroot=true;
					for (j=0;j<nmin;j++){
						if (fabs(xmin1-amin[j]) <= (EQL*xmin1)) newroot=false;
						if (newroot) {
							amin[nmin++]=xmin1;
						}
					}
				} 
				//end solve 

				tmp_eval_vl = eval_f_vl2_npent(xmin1);
				
				tmp_x_m12 = xmin1;
				tmp_x_m11 = ( out_m_k[2]/beta - pi_2*tmp_x_m12 ) / pi_1;
				
				if ( tmp_eval_vl >= vh_v0
					&& tmp_eval_vl <= vh_v1 + 0.0
					//&& tmp_x_m11 >= (m1_s1_l[0]+m1_s1_r[0])/2.0
					//&& tmp_x_m11 <= (m1_s1_l[Nm-1]+m1_s1_r[Nm-1])/2.0
					){   
						
						tmp_vl2 = tmp_eval_vl;
				}
				//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

				tmp_W_maxmin[i_tau1][i_tau2] = min_dp(tmp_vl1, tmp_vl2);
				
				//tmp_vl = max_dp(tmp_vl1, tmp_vl2);
				if (min_dp(tmp_vl1, tmp_vl2) < tmp_vl){
					tmp_vl = min_dp(tmp_vl1, tmp_vl2);
				}

				//if (min_dp(tmp_vl1, tmp_vl2) >= vh_v0){
				//}
			
			}//end if.. 
		}// end tau2 
	}//end tau1

	//return tmp_vh;  
	renew_v[0] = tmp_vl;
	renew_v[1] = tmp_vh;
}

double intp_1d(double &x_k, int &shk, double policy[Nk][Ns])
{   
	int i_k;	
	NRlocate(k_vec, Nk, x_k, i_k);
    if (i_k==-1){i_k = 0;}
	if (i_k==Nk-1){i_k = Nk-2;}
		
	return policy[i_k][shk] + (policy[i_k+1][shk]-policy[i_k][shk])
		/(k_vec[i_k+1] - k_vec[i_k])*(x_k - k_vec[i_k]);	
}   

void print_Wm( )
{
    //static
	ofstream file("policy_Wm_sto_2013_3.txt");
	file.precision(4); 
	    
	for (int i_s = 0; i_s < Ns; i_s++) {
		for (int row = 0; row < Nk; row ++) {
		
		file<<k_vec[row]<<" "<<gc_1[row][i_s]<<" "<<gl_1[row][i_s]<<" "<<gk1_1[row][i_s]<<" "
		    <<gtauk_1[row][i_s]<<" "<<gtaul_1[row][i_s]<<" "<<Wv0_0[row][i_s][0]<<endl;
		}
	}  
}
 
void input_Wm( )  // import the initial endowment matrix.
{
    //int k,l;
	string txt;
    ifstream fp("policy_Wm_sto_7_5.txt");
       
	for (int i_s = 0; i_s < Ns; i_s++) {
		for (int row = 0; row < Nk; row ++) {
			double dg1, dg2, dg3;
			fp>>dg1>>dg2>>dg3;   
			
			Wm_0[row][i_s][0] = dg1 - 0.0;
			Wm_0[row][i_s][1] = dg2;
			
		}
	}
	
   fp.close();        
}

void input_Wv( )
{  
	string txt;
    ifstream fp("policy_Wv_sto_2013_4.txt");
	
	for (int i_s = 0; i_s < Ns; i_s++) {
		for (int row = 0; row < Nk; row ++) {
			for (int i_m = 0; i_m < Nm; i_m ++) {   
				
				double dg1, dg2, dg3, dg4; 

				fp>>dg1>>dg2>>dg3>>dg4;   

                         
				if (i_m == 0){
					Wm_0[row][i_s][0] = dg2 - 0.005;
									}
				if (i_m == Nm-1){
					Wm_0[row][i_s][1] = dg2 + 0.005;
				}
				
				if (dg3 < v_lower){
					dg3 = v_lower;
				}
				Wv0_0[row][i_s][i_m] = dg3 - 0.01; //
				Wv_0[row][i_s][i_m] = dg4 + 0.01; //

			}
		}
	}
}

void print_Wv( )
{  
	ofstream file("policy_Wv_sto_2013_4-1.txt");
	file.precision(6);
	double tmp_m0;
	for (int i_s = 0; i_s < Ns; i_s++) {
		for (int row = 0; row < Nk; row ++) {
			for (int i_m = 0; i_m < Nm; i_m ++) {
				tmp_m0 = Wm_1[row][i_s][0] + (Wm_1[row][i_s][1] - Wm_1[row][i_s][0]) * i_m / (Nm-1); 
    			file <<k_vec[row]<<" "<<tmp_m0<<" "<<Wv0_1[row][i_s][i_m]<<" "<<Wv_1[row][i_s][i_m]<<endl;
			}
		}
	}
}

void print_Wv_short( )
{  
	ofstream file("policy_Wv_short_sto_2013_3.txt");
	file.precision(4);
	double tmp_m0;
	for (int i_s = 0; i_s < Ns; i_s++) {
		for (int row = 50; row < Nk; row += 20) {
			for (int i_m = 0; i_m < Nm; i_m += 2) {
				tmp_m0 = Wm_1[row][i_s][0] + (Wm_1[row][i_s][1] - Wm_1[row][i_s][0]) * i_m / (Nm-1); 
    			file <<k_vec[row]<<" "<<tmp_m0<<" "<<Wv0_1[row][i_s][i_m]<<" "<<Wv_1[row][i_s][i_m]<<endl;
			}
		}
	}
}

int main(int argc, char** argv)
{
      ofstream fout;
      fout.open("game_sto_sse.txt");
      fout.setf(ios::fixed);
      fout.precision(6);

	  p11 = 0.946;
	  p12 = 0.054;
	  p21 = 0.054;
      p22 = 0.946;

		// initialization of MPI related constants
	    int rank;        /* Rank of process */
	    int n_proc;     /* Number of process */
	    int root = 0;    /* Define root */
	    int stat = MPI_SUCCESS;   /* Defind status*/

		double s_time, e_time;

	    MPI_Status status; /* Return status for MPI_BCAST and MPI_GATHER */
	
		MPI_Init(&argc, &argv); /* Initialization */
	    MPI_Comm_rank(MPI_COMM_WORLD, &rank); /* Get rank */
	    MPI_Comm_size(MPI_COMM_WORLD, &n_proc); /* Get size */
	
		const int N_msg = Nk*Ns*(2*Nm+2 + 5)/N_cpu; //5: gm0, gm1, gk, gtauk, gtaul   
		const int N_array = Nk*Ns*(2*Nm+2 + 5);
		double msg[N_msg];// each process allocate maximum certain amount of memory for this array
		double msg_Array[N_array];
	
		//***********************************************************//
		for (int row = 0; row < Nk; row++) {
            k_vec[row] = (kh-kl)/(Nk-1)*(row)+kl;
		}
		
		for (int row = 0; row < Ntau; row++) {
	        tau_vec[row] = (tau_b-tau_a)/(Ntau-1)*(row)+tau_a;
		} 
        
		//input_Wm( );
		input_Wv( );

		for (int row = 0; row < Nk; row++) {
			for (int i_s = 0; i_s < Ns; i_s++) {
				W_pent[row][i_s] = 100.0;

    			for (int i_m = 0; i_m < Nm; i_m++) {

					if (Wv0_0[row][i_s][i_m] < W_pent[row][i_s])
					{
						W_pent[row][i_s] =  Wv0_0[row][i_s][i_m];
					} 

				}
			}
		}
       

	//initialize W1:
		for (int row = 0; row < Nk; row++) {
			for (int i_s = 0; i_s < Ns; i_s++) {
				for (int i_m = 0; i_m < Nbdr; i_m++) {
					Wm_1[row][i_s][i_m] = Wm_0[row][i_s][i_m];
				}
				
				gc_1[row][i_s] = k_vec[row];
				gk1_1[row][i_s] = k_vec[row];
				gl_1[row][i_s] = 0.3;
				gtauk_1[row][i_s] = 0.85;
				gtaul_1[row][i_s] = 0.1;
			}
		}
		for (int row = 0; row < Nk; row++) {
			for (int i_s = 0; i_s < Ns; i_s++) {
				//Wv0_1[row][i_s] = Wv0_0[row][i_s];

				for (int i_m = 0; i_m < Nm; i_m++) { 
					Wv0_1[row][i_s][i_m] = Wv0_0[row][i_s][i_m];
					
					Wv_1[row][i_s][i_m] = Wv_0[row][i_s][i_m];
				}
			}
		}

	int iter, iter_dist, iter_update_v, iter_print, iter_stop;
	int iter_update_vl;
	iter = 0;
	iter_dist = 0;
	tol_v0 = 0.001 / 2.0;
	iter_update_v = 0;
	iter_print = 0;
	iter_stop = 0;
	iter_update_vl = 0;

	do {
		iter += 1;
		//renew the grid for W.
		for (int row = 0; row < Nk; row++) {
			for (int i_s = 0; i_s < Ns; i_s++) {
			
				gc_0[row][i_s] = gc_1[row][i_s];
				gl_0[row][i_s] = gl_1[row][i_s];
				gk1_0[row][i_s] = gk1_1[row][i_s];
				gtauk_0[row][i_s] = gtauk_1[row][i_s];
				gtaul_0[row][i_s] = gtaul_1[row][i_s];
				
				for (int i_m = 0; i_m < Nbdr; i_m++) {
					Wm_0[row][i_s][i_m] = Wm_1[row][i_s][i_m];
				}
			}
		}
		for (int row = 0; row < Nk; row++) {
			for (int i_s = 0; i_s < Ns; i_s++) {
				
				for (int i_m = 0; i_m < Nm; i_m++) {
					Wv0_0[row][i_s][i_m] = Wv0_1[row][i_s][i_m]; 
					
					Wv_0[row][i_s][i_m] = Wv_1[row][i_s][i_m];
				}
			}
		}  
		//punishment = min{v(k,m)}
   		for (int i_s = 0; i_s < Ns; i_s++) {
			for (int row = 0; row < Nk; row ++) {     
				
				double tmp_pent = 100;
				     
				for (int i_m = 0; i_m < Nm; i_m ++) {   
                    
					if (Wv0_0[row][i_s][i_m] < tmp_pent)
					{
						tmp_pent =  Wv0_0[row][i_s][i_m];
					}   
				}  
				
				W_pent[row][i_s] = tmp_pent;
			}
		}     
		//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
		
		fout << iter<< endl;  		
		for (int row = 0; row < Nk/n_proc; row++) {
			
			k0 = k_vec[rank*Nk/n_proc + row];

			for (int i_s = 0; i_s < Ns; i_s++) {	

				//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
				//update Wm:
				for (int i_m = 0; i_m < Nbdr; i_m++){
					
					//check_W = 0;
					if (i_m == 0){
						msg[(row*Ns+i_s) * (2*Nm+2 + 5) + 0] = Wm_0[rank*Nk/n_proc + row][i_s][i_m];
					}
					if (i_m == 1){
						msg[(row*Ns+i_s) * (2*Nm+2 + 5) + 1] = Wm_0[rank*Nk/n_proc + row][i_s][i_m];
					}

						shock(i_s, a0, pi_1, pi_2);

						m0 = Wm_0[rank*Nk/n_proc + row][i_s][i_m];

						check_equ(k0, i_s, m0);

						check_m = check_vec[0];
						check_v = check_vec[1];

						if (check_m == 0){  //if check does not find suitable m+, v, then move the bdr.
							//lower bound:
							if (i_m == 0){  
								msg[(row*Ns+i_s) * (2*Nm+2 + 5) + 0] = Wm_0[rank*Nk/n_proc + row][i_s][i_m]
									+ tol_v0;  
							}
							//upper bound:
							if (i_m == 1){
								msg[(row*Ns+i_s) * (2*Nm+2 + 5) + 1] = Wm_0[rank*Nk/n_proc + row][i_s][i_m] 
									- tol_v0;
							}							
						}// end check_m
				}//end i_m

				//the following is used to prevent the case that one crossing will stop the algorithm...06.25.2010
				//if m_l > m_h, then m_l = m_h = (m_l + m_h)/2.0;  
				if ( msg[(row*Ns+i_s) * (2*Nm+2 + 5) + 0] >= msg[(row*Ns+i_s) * (2*Nm+2 + 5) + 1] ) 
				{ 
					double tmp_msg = ( msg[(row*Ns+i_s) * (2*Nm+2 + 5) + 0] + msg[(row*Ns+i_s) * (2*Nm+2 + 5) + 1] ) / 2.0;  
					msg[(row*Ns+i_s) * (2*Nm+2 + 5) + 0] = tmp_msg;
					msg[(row*Ns+i_s) * (2*Nm+2 + 5) + 1] = tmp_msg;
				}//end if crossing.

				for (int i_m_v = 0; i_m_v < Nm; i_m_v++){

					msg[(row*Ns+i_s) * (2*Nm+2 + 5) + 2 + i_m_v] = Wv0_0[rank*Nk/n_proc + row][i_s][i_m_v]; 
					msg[(row*Ns+i_s) * (2*Nm+2 + 5) + 2 + Nm + i_m_v] = Wv_0[rank*Nk/n_proc + row][i_s][i_m_v]; 
				}
			}//end i_s
		}//end row
		
				MPI_Barrier(MPI_COMM_WORLD);
				stat = MPI_Gather(msg, N_msg, MPI_DOUBLE, msg_Array, N_msg, MPI_DOUBLE, root, MPI_COMM_WORLD);
		
				if (rank == root){
					for (int row = 0; row < Nk; row++) {
						for (int i_s = 0; i_s < Ns; i_s++) {
							Wm_1[row][i_s][0] = msg_Array[(row*Ns+i_s) * (2*Nm+2 + 5) + 0];
							Wm_1[row][i_s][1] = msg_Array[(row*Ns+i_s) * (2*Nm+2 + 5) + 1];
						}
					}
					
					//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
					//dynamic adjustment
					double adjustment_grid[Nm], adjustment_value[Nm];
					double tmp_adjust_m0;
					double tmp_adjust_vh[Nk][Ns][Nm];
					int adjust_lef;
					
					if ( iter_update_v > 1) {
						for (int row = 0; row < Nk; row++) {
							for (int i_s = 0; i_s < Ns; i_s++) {

								for (int i_m_v = 0; i_m_v < Nm; i_m_v++){
									tmp_adjust_vh[row][i_s][i_m_v] = Wv_0[row][i_s][i_m_v];	
								
								}
								
								if ( fabs (Wm_0[row][i_s][0] - Wm_1[row][i_s][0]) > tol_v0/2 
									 ||	fabs (Wm_0[row][i_s][1] - Wm_1[row][i_s][1]) > tol_v0/2 ){

									//the grid is given by: Wm_0 --> Wv_0 
									for (int i_m_v = 0; i_m_v < Nm; i_m_v++){
										adjustment_grid[i_m_v] = Wm_0[row][i_s][0] + (Wm_0[row][i_s][1] - Wm_0[row][i_s][0]) *i_m_v/ (Nm-1);
										adjustment_value[i_m_v] = Wv_0[row][i_s][i_m_v];
									}
								
									//dynamic adjust the new grid: Wm_1 --> Wv_1
									for (int i_m_v = 0; i_m_v < Nm; i_m_v++){
										
										tmp_adjust_m0 = Wm_1[row][i_s][0] + (Wm_1[row][i_s][1] - Wm_1[row][i_s][0]) *i_m_v/ (Nm-1);
									
										NRlocate(adjustment_grid, Nm, tmp_adjust_m0, adjust_lef);
									    if (adjust_lef==-1){adjust_lef = 0;}
										if (adjust_lef==Nm-1){adjust_lef = Nm-2;}
									
										tmp_adjust_vh[row][i_s][i_m_v] = adjustment_value[adjust_lef] + (adjustment_value[adjust_lef+1] - adjustment_value[adjust_lef])
											/(adjustment_grid[adjust_lef+1] - adjustment_grid[adjust_lef])*(tmp_adjust_m0 - adjustment_grid[adjust_lef]);
											
										Wv_0[row][i_s][i_m_v] = tmp_adjust_vh[row][i_s][i_m_v];		
									
									}
								}//end if wm_0 <> wm_1
							}
						}

					}//end if
					//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

				}//end if root == 0

				e_time = MPI_Wtime();
				stat = MPI_Bcast (Wm_1, Nk*Ns*Nbdr, MPI_DOUBLE, root, MPI_COMM_WORLD);// bcast W
				stat = MPI_Bcast (Wv0_0, Nk*Ns*Nm, MPI_DOUBLE, root, MPI_COMM_WORLD);// bcast W
				stat = MPI_Bcast (Wv_0, Nk*Ns*Nm, MPI_DOUBLE, root, MPI_COMM_WORLD);// bcast W
				//end update Wm.
				//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
				
				//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
				//update Wv:
				
				check_dis(); 
				
				double tmp_vl, tmp_vh, tmp_vh_min;//, vec_vh_min[Nk/n_proc];

				if ( iter > 1 //&& iter_update_v >= 1 ) {       //test, it does not work.
					 && max_dis_m <= tol_v0 / 2.00) {

					for (int row = 0; row < Nk/n_proc; row++) {
			
						k0 = k_vec[rank*Nk/n_proc + row];

						for (int i_s = 0; i_s < Ns; i_s++) {	

							//update Wv
							
							tmp_vh_min = 100.0;
							
							msg[(row*Ns+i_s) * (2*Nm+2 + 5) + 2 + Nm*2 + 0] = gc_0[row][i_s];
							msg[(row*Ns+i_s) * (2*Nm+2 + 5) + 2 + Nm*2 + 1] = gk1_0[row][i_s];
							msg[(row*Ns+i_s) * (2*Nm+2 + 5) + 2 + Nm*2 + 2] = gtauk_0[row][i_s];
							msg[(row*Ns+i_s) * (2*Nm+2 + 5) + 2 + Nm*2 + 3] = gtaul_0[row][i_s];
							msg[(row*Ns+i_s) * (2*Nm+2 + 5) + 2 + Nm*2 + 4] = gl_0[row][i_s];
							

							for (int i_m_v = 0; i_m_v < Nm; i_m_v++){

								msg[(row*Ns+i_s) * (2*Nm+2 + 5) + 2 + i_m_v] = Wv0_0[rank*Nk/n_proc + row][i_s][i_m_v]; 
								msg[(row*Ns+i_s) * (2*Nm+2 + 5) + 2 + Nm + i_m_v] = Wv_0[rank*Nk/n_proc + row][i_s][i_m_v]; 
								
								if (Wv_0[rank*Nk/n_proc + row][i_s][i_m_v] <= tmp_vh_min 
									&& Wv_0[rank*Nk/n_proc + row][i_s][i_m_v] > v_lower){
									tmp_vh_min = Wv_0[rank*Nk/n_proc + row][i_s][i_m_v];
								}

								if ( Wm_1[rank*Nk/n_proc + row][i_s][0] <= Wm_1[rank*Nk/n_proc + row][i_s][1]) { 
		
									if (i_m_v == 0){
										msg[(row*Ns+i_s) * (2*Nm+2 + 5) + 0] = Wm_1[rank*Nk/n_proc + row][i_s][0];
									}
									if (i_m_v == Nm-1){
										msg[(row*Ns+i_s) * (2*Nm+2 + 5) + 1] = Wm_1[rank*Nk/n_proc + row][i_s][1];
									}						
		
									m0 = Wm_1[rank*Nk/n_proc + row][i_s][0] + 
										(Wm_1[rank*Nk/n_proc + row][i_s][1] - Wm_1[rank*Nk/n_proc + row][i_s][0]) *i_m_v/ (Nm-1);
	
									update_vh(k0, m0, i_s);   

									for (int i_tau1 = 0; i_tau1 < Ntau; i_tau1 ++){
										for (int i_tau2 = 0; i_tau2 < Ntau; i_tau2 ++){
											W_maxmin[i_tau1][i_tau2][i_m_v] = tmp_W_maxmin[i_tau1][i_tau2];
											k1_maxmin[i_tau1][i_tau2][i_m_v] = tmp_k1_maxmin[i_tau1][i_tau2];
											c_maxmin[i_tau1][i_tau2][i_m_v] = tmp_c_maxmin[i_tau1][i_tau2];
											l_maxmin[i_tau1][i_tau2][i_m_v] = tmp_l_maxmin[i_tau1][i_tau2];
										}
									}
				           
									tmp_vl = renew_v[0]; 
									tmp_vh = renew_v[1];  
					
									//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
									//update vl first, then vh.    
									if(iter_update_vl < 1){
										tmp_vh_min = v_upper;	
									}//end iter_update_vl
					
									if(iter_update_vl > 1){
					 
										if ( tmp_vh >= Wv0_0[rank*Nk/n_proc + row][i_s][i_m_v] - 0.0
											&& tmp_vh <= v_upper  // Wv_0[rank*Nk/n_proc + row][i_s][i_m_v] + 0.0 
										){
										
											msg[(row*Ns+i_s) * (2*Nm+2 + 5) + 2 + Nm + i_m_v] = tmp_vh;
										}  
						
										if (tmp_vh <= tmp_vh_min && tmp_vh > v_lower){
											tmp_vh_min = tmp_vh;
										}              
									}//end iter_update_vl
									//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
									
								}//end Wm[0] <= Wm[1].
							
							//vec_vh_min[rank*Nk/n_proc + row] = tmp_vh_min;

						}// end i_m_v.

				//find the max_min_{u+beta*v}
				double tmp_pent_max, tmp_pent_min;
				tmp_pent_max = -10.0;
				
				double tmp_c_max, tmp_c_min, tmp_l_max, tmp_l_min;
				tmp_c_max = -10.0;
				tmp_l_max = -10.0;

				double tmp_k1_max, tmp_k1_min;
				tmp_k1_max = -10.0;

				double tauk_min, tauk_max, taul_min, taul_max;
				
				tauk_max = tau_b;
				taul_max = tau_b;
				//tmp_k1_max = 1.0;
				//tmp_m_max = 1.0; 
				
				for (int i_tau1 = 0; i_tau1 < Ntau; i_tau1 ++){
					for (int i_tau2 = 0; i_tau2 < Ntau; i_tau2 ++){

						tmp_pent_min = 10.0;
						tmp_c_min = 10.0;
						tmp_l_min = 10.0;
						tmp_k1_min = 10.0;

						tauk_min = tau_vec[i_tau1];
						taul_min = tau_vec[i_tau2];

						for (int i_m_v = 0; i_m_v < Nm; i_m_v++){
							if (											
								W_maxmin[i_tau1][i_tau2][i_m_v] <= tmp_pent_min 											
								&& W_maxmin[i_tau1][i_tau2][i_m_v] >= Wv0_0[rank*Nk/n_proc + row][i_s][i_m_v]    //v_lower
								&& W_maxmin[i_tau1][i_tau2][i_m_v] < v_upper											
								&& W_maxmin[i_tau1][i_tau2][i_m_v] <= tmp_vh_min
							){
								tmp_pent_min = W_maxmin[i_tau1][i_tau2][i_m_v];
								tmp_k1_min = k1_maxmin[i_tau1][i_tau2][i_m_v];								
								tmp_c_min = c_maxmin[i_tau1][i_tau2][i_m_v];
								tmp_l_min = l_maxmin[i_tau1][i_tau2][i_m_v];
							}//end if
						}//end i_m_v
						//fout<<W_maxmin[i_tau2][3]<<endl;

						if ( tmp_pent_min >= Wv0_0[rank*Nk/n_proc + row][i_s][0] // v_lower 
							&& tmp_pent_min < v_upper 
							&& tmp_pent_min >= tmp_pent_max){
							
							tmp_pent_max = tmp_pent_min;
							tmp_k1_max = tmp_k1_min;
							tmp_c_max = tmp_c_min;
							tmp_l_max = tmp_l_min;

							tauk_max = tauk_min;
							taul_max = taul_min;
						}
					}
				}

				msg[(row*Ns+i_s) * (2*Nm+2 + 5) + 2 + Nm*2 + 0] = tmp_c_max;
				msg[(row*Ns+i_s) * (2*Nm+2 + 5) + 2 + Nm*2 + 1] = tmp_k1_max;
				msg[(row*Ns+i_s) * (2*Nm+2 + 5) + 2 + Nm*2 + 2] = tauk_max;
				msg[(row*Ns+i_s) * (2*Nm+2 + 5) + 2 + Nm*2 + 3] = taul_max;
				msg[(row*Ns+i_s) * (2*Nm+2 + 5) + 2 + Nm*2 + 4] = tmp_l_max;

				for (int i_m_v = 0; i_m_v < Nm; i_m_v++){
					if ( //tmp_pent_max > v_lower  && tmp_pent_max < tmp_vh_min){
					tmp_pent_max >= msg[(row*Ns+i_s) * (2*Nm+2 + 5) + 2 + i_m_v] 
					 && tmp_pent_max <= msg[(row*Ns+i_s) * (2*Nm+2 + 5) + 2 + Nm + i_m_v] ) {

						msg[(row*Ns+i_s) * (2*Nm+2 + 5) + 2 + i_m_v] = tmp_pent_max;
						

					}//end if
				}//end i_m_v
	
				//the following is used to prevent the case that one crossing will stop the algorithm...06.25.2010
				//if m_l > m_h, then m_l = m_h = (m_l + m_h)/2.0;  
				if ( msg[(row*Ns+i_s) * (2*Nm+2 + 5) + 0] >= msg[(row*Ns+i_s) * (2*Nm+2 + 5) + 1] ) 
				{ 
					double tmp_msg = ( msg[(row*Ns+i_s) * (2*Nm+2 + 5) + 0] + msg[(row*Ns+i_s) * (2*Nm+2 + 5) + 1] ) / 2.0;  
					msg[(row*Ns+i_s) * (2*Nm+2 + 5) + 0] = tmp_msg;
					msg[(row*Ns+i_s) * (2*Nm+2 + 5) + 1] = tmp_msg;
				}//end if crossing.

			}// end i_s
			}// end row
		}// end if.
		//end update Wv
				//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
		
		MPI_Barrier(MPI_COMM_WORLD);
		stat = MPI_Gather(msg, N_msg, MPI_DOUBLE, msg_Array, N_msg, MPI_DOUBLE, root, MPI_COMM_WORLD);
		
		if (rank == root){
		    for (int row = 0; row < Nk; row++) {
				for (int i_s = 0; i_s < Ns; i_s++) {
					Wm_1[row][i_s][0] = msg_Array[(row*Ns+i_s) * (2*Nm+2 + 5) + 0];
					Wm_1[row][i_s][1] = msg_Array[(row*Ns+i_s) * (2*Nm+2 + 5) + 1];
					
					gc_1[row][i_s] = msg_Array[(row*Ns+i_s) * (2*Nm+2 + 5) + 2 + Nm*2 + 0];
					gk1_1[row][i_s] = msg_Array[(row*Ns+i_s) * (2*Nm+2 + 5) + 2 + Nm*2 + 1];
					gtauk_1[row][i_s] = msg_Array[(row*Ns+i_s) * (2*Nm+2 + 5) + 2 + Nm*2 + 2];
					gtaul_1[row][i_s] = msg_Array[(row*Ns+i_s) * (2*Nm+2 + 5) + 2 + Nm*2 + 3];
					gl_1[row][i_s] = msg_Array[(row*Ns+i_s) * (2*Nm+2 + 5) + 2 + Nm*2 + 4];
				}
			}

			for (int row = 0; row < Nk; row++) {
				for (int i_s = 0; i_s < Ns; i_s++) {
					for (int i_m = 0; i_m < Nm; i_m++) {     
						Wv0_1[row][i_s][i_m] = msg_Array[(row*Ns+i_s) * (2*Nm+2 + 5) + 2 + i_m];
						
						Wv_1[row][i_s][i_m] = msg_Array[(row*Ns+i_s) * (2*Nm+2 + 5) + 2 + Nm + i_m];
					}
				}
			}  
		}//end if root == 0
		e_time = MPI_Wtime();
		stat = MPI_Bcast (Wm_1, Nk*Ns*Nbdr, MPI_DOUBLE, root, MPI_COMM_WORLD);// bcast W
		stat = MPI_Bcast (Wv0_1, Nk*Ns*Nm, MPI_DOUBLE, root, MPI_COMM_WORLD);// bcast W
		stat = MPI_Bcast (Wv_1, Nk*Ns*Nm, MPI_DOUBLE, root, MPI_COMM_WORLD);// bcast W
		//+++++++++++++++++++++++++++++++++++
		stat = MPI_Bcast (gc_1, Nk*Ns, MPI_DOUBLE, root, MPI_COMM_WORLD);// bcast W
		stat = MPI_Bcast (gk1_1, Nk*Ns, MPI_DOUBLE, root, MPI_COMM_WORLD);// bcast W
		stat = MPI_Bcast (gl_1, Nk*Ns, MPI_DOUBLE, root, MPI_COMM_WORLD);// bcast W
		stat = MPI_Bcast (gtauk_1, Nk*Ns, MPI_DOUBLE, root, MPI_COMM_WORLD);// bcast W
		stat = MPI_Bcast (gtaul_1, Nk*Ns, MPI_DOUBLE, root, MPI_COMM_WORLD);// bcast W

		if (stat != MPI_SUCCESS )
		{
			fout<<"FATAL ERROR: "<<stat<<endl;
			//exit(1);
			break;
		}
		
		if (iter_print == 0) {
			if (root == 0) { 
				print_Wv_short();
				print_Wm();
				print_Wv();
				
			}  //end: if  root==0
		}  //end if iter==??.


		check_dis(); 

		//add Wv ........
		double iter_Wv;
		iter_Wv = 0;
		if (max_dis_m>=tol_v0 && max_dis_vl>=tol_v0 && max_dis_vh>=tol_v0){
			iter_Wv += 1.0;
			
			for (int row = 0; row < Nk; row++) {
				for (int i_s = 0; i_s < Ns; i_s++) {	
					for (int i_m = 0; i_m < Nm; i_m++) {
						Wv_1[row][i_s][i_m] = Wv_1[row][i_s][i_m] + 0.01 * pow(beta, iter_Wv);
						
						//Wv0_1[row][i_s][i_m] = Wv0_1[row][i_s][i_m] - 0.01 * pow(beta, iter_Wv);
					}
				}//end i_s
			}//end row
		}

		fout<<max_dis_m<<" "<<max_dis_vl<<" "<<max_dis_vh<<endl;
		
		//step 1. update m
		if (max_dis_m <= tol_v0/2.0){
			iter_update_v += 1;
		}
		//step 2. update vl.
		if (max_dis_m < tol_v0/2.0 && max_dis_vl <= 0.5){
			iter_update_vl += 1;
		}
		//step 3. update vh.
		if (max_dis_m < tol_v0/2.0 && max_dis_vl <= tol_v0 && max_dis_vh <= tol_v0){
			iter_dist += 1;
		}

		iter_print += 1;

		if (iter_print >= 9){
			iter_print = 0;
		}

	} while (iter == 1 || iter_dist < 50 //|| iter_stop < 1 //iter_dist < 300 
		     || max_dis_m > tol_v0 || max_dis_m > tol_v0 || max_dis_vh > tol_v0); //(!check());
	
	fout << " End of Iterations!" << endl;
	fout << "*******************************************************" << endl; 
	
	if (root == 0) { 
		//print Wm, print Wv
		print_Wm();
		print_Wv(); 
		print_Wv_short();
	}  //end: if
	

	MPI_Finalize();
    return 0;
}  
